/*================================================================*/
/* COPYRIGHT:                                                     */
/* ----------                                                     */
/* Copyright (C) International Business Machines Corp., 1996.     */
/*                                                                */
/* DISCLAIMER OF WARRANTIES:                                      */
/* -------------------------                                      */
/* The following [enclosed] code is sample code created by IBM    */
/* Corporation.  This sample code is not part of any standard IBM */
/* product and is provided to you solely for the purpose of       */
/* assisting you in the development of your applications.  The    */
/* code is provided "AS IS", without warranty of any kind.  IBM   */
/* shall not be liable for any damages arising out of your use of */
/* the sample code, even if they have been advised of the         */
/* possibility of such damages.                                   */
/*================================================================*/

#define STRICT
#include <windows.h>
#include <io.h>

#include <stdio.h>
#include <stdlib.h>


#include <fcntl.h>
#include <sys/stat.h>


#include "audwav.h"
#include "waves.h"


#ifdef BIGENDIAN
#define SwapShort(xx)  _SwapShort(xx)
#define SwapLong(xx)   _SwapLong(xx)
#else
#define SwapShort(xx)  ( xx )
#define SwapLong(xx)   ( xx )
#endif


extern FILE * the_log_file;
extern int    the_log_level;
       char   log_buff [ 255 ];


static int wav_data_pos = 0;
static int wav_data_len = 0;


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
unsigned short _SwapShort ( short data )
{
  short lo;
  short hi;

  lo = ( data & 0x00ff );
  hi = ( data & 0xff00 );

  return ( ( lo << 8 ) | ( hi >> 8 ) );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
unsigned long _SwapLong ( unsigned long data )
{
  unsigned short lo;
  unsigned short hi;
  unsigned short rlo;
  unsigned short rhi;

  lo = ( unsigned short ) ( data & 0x0000ffff );
  hi = ( unsigned short ) ( ( data & 0xffff0000 ) >> 16 );

  rlo = SwapShort ( lo );
  rhi = SwapShort ( hi );

  return ( ( rlo << 16 ) | ( rhi ) );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
static int NextShort ( int fd, unsigned short * stemp )
{
  int rc;

  rc = read ( fd, stemp, sizeof ( * stemp ) );

  if ( rc != sizeof ( * stemp ) )
  {
    return ( -1 );
  }

  return ( 0 );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
static int NextLong ( int fd, unsigned long * ltemp )
{
  int rc;

  rc = read ( fd, ltemp, sizeof ( * ltemp ) );

  if ( rc != sizeof ( * ltemp ) )
  {
    return ( -1 );
  }

  return ( 0 );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
int WavReadFmt ( int fd, WaveFormatStruct * fmt )
{
  int            rc;
  unsigned long  ltemp;
  unsigned short stemp;
  short          format_len;

  /* Assume that we are pointing at the first byte PAST the 'fmt ' string */

  if ( NextLong ( fd, & ltemp ) )                    /* Length of section */
  {
    sprintf ( log_buff, "WavReadFmt: read(length) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  format_len = ( short ) SwapLong ( ltemp );


  if ( NextShort ( fd, & stemp ) )                        /* Data format */
  {
    sprintf ( log_buff, "WavReadFmt: read(format) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  fmt->format = SwapShort ( stemp );


  if ( NextShort ( fd, & stemp ) )                       /* Num Channels */
  {
    sprintf ( log_buff, "WavReadFmt: read(channels) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  fmt->num_channels = SwapShort ( stemp );


  if ( NextLong ( fd, & ltemp ) )                     /* Samples per sec */
  {
    sprintf ( log_buff, "WavReadFmt: read(samples/s) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  fmt->samples_per_sec = SwapLong ( ltemp );


  if ( NextLong ( fd, & ltemp ) )                       /* Bytes per sec */
  {
    sprintf ( log_buff, "WavReadFmt: read(bytes/s) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  fmt->avg_bytes_per_sec = SwapLong ( ltemp );


  if ( NextShort ( fd, & stemp ) )                        /* Block Align */
  {
    sprintf ( log_buff, "WavReadFmt: read(blk align) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  fmt->block_align = SwapShort ( stemp );


  if ( NextShort ( fd, & stemp ) )                    /* Bits per sample */
  {
    sprintf ( log_buff, "WavReadFmt: read(bits/smpl) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  fmt->bits_per_sample = SwapShort ( stemp );


  /* If this an extended one, the length is 18 (a word at the end for */
  /* the 'extra' data)                                                */
  if ( format_len != 16 )
  {
    if ( NextShort ( fd, & stemp ) )
    {
      sprintf ( log_buff, "WavReadFmt: read(extra) failed, errno=%d\n", errno );
      ErrorMsg ( log_buff );

      return ( -1 );
    }

    fmt->extra = SwapShort ( stemp );

    /* if there is 'extra' and this file is NOT a 'WAVE_FORMAT_PCM' file */
    /* then we skip over the 'extra' data (this tidbit comes from the    */
    /* MSDEV description of the 'WAVEFORMATEX' structure                 */
    if ( ( fmt->extra ) && ( fmt->format != 0x01 ) )
    {
      if ( ( rc = lseek ( fd, fmt->extra, SEEK_CUR ) ) == -1 )
      {
        sprintf ( log_buff, "WavReadFmt: lseek failed, errno=%d\n", errno );
        ErrorMsg ( log_buff );

        return ( -1 );
      }
    }
  }

  if ( the_log_level )
  {
    sprintf ( log_buff, "WavReadFmt: Length = %d\n", format_len );
    LogMsg ( log_buff );
    sprintf ( log_buff, "WavReadFmt: Format = %d\n", fmt->format );
    LogMsg ( log_buff );
    sprintf ( log_buff, "WavReadFmt: Channels = %d\n", fmt->num_channels );
    LogMsg ( log_buff );
    sprintf ( log_buff, "WavReadFmt: Samples/sec = %d\n", fmt->samples_per_sec );
    LogMsg ( log_buff );
    sprintf ( log_buff, "WavReadFmt: Bytes/sec = %d\n", fmt->avg_bytes_per_sec );
    LogMsg ( log_buff );
    sprintf ( log_buff, "WavReadFmt: Align = %d\n", fmt->block_align );
    LogMsg ( log_buff );
    sprintf ( log_buff, "WavReadFmt: Bits/Sample = %d\n", fmt->bits_per_sample );
    LogMsg ( log_buff );

    if ( fmt->format == 0x01 )
    {
      sprintf ( log_buff, "WavReadFmt: Extra = %d (ignored)\n", fmt->extra );
    }
    else
    {
      sprintf ( log_buff, "WavReadFmt: Extra = %d\n", fmt->extra );
    }
    LogMsg ( log_buff );
  }

  return ( 0 );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
int WavReadData ( int fd )
{
  int           rc;
  unsigned long ltemp;
  long          data_len;

  /* Assume that we are pointing at the first byte PAST the 'data' string */

  if ( NextLong ( fd, & ltemp ) )
  {
    sprintf ( log_buff, "WavReadData: read(length) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  data_len = SwapLong ( ltemp );


  if ( ( rc = lseek ( fd, data_len, SEEK_CUR ) ) == -1 )
  {
    sprintf ( log_buff, "WavReadData: lseek failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }


  sprintf ( log_buff, "WavReadData: Length = %d\n", data_len );
  LogMsg ( log_buff );

  return ( 0 );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
int WavNextBlock ( int fd, char * stype )
{
  int           rc;
  unsigned long ltemp;
  long          sect_len;
  char          sect_type [ 4 ];

  /* Assume we are pointing at the character string at the start of a */
  /* section (any one in the file)                                    */

  /* Go through each of the 'blocks' that are in the file, until we find */
  /* one that matches the type we are looking for                        */

  while ( NextLong ( fd, ( unsigned long * ) sect_type ) == 0 )
  {
    /* If the shoe fits, return it */
    if ( strncmp ( sect_type, stype, 4 ) == 0 )
    {
      return ( 0 );
    }

    /* If this isn't the right section type, then figure out how long it */
    /* is and skip past it..                                             */
    if ( NextLong ( fd, & ltemp ) )
    {
      sprintf ( log_buff, "WavNextBlock: read(len) failed, errno=%d\n", errno );
      ErrorMsg ( log_buff );

      return ( -1 );
    }

    sect_len = SwapLong ( ltemp );

    if ( ( rc = lseek ( fd, sect_len, SEEK_CUR ) ) == -1 )
    {
      sprintf ( log_buff, "WavNextBlock: lseek failed, errno=%d\n", errno );
      ErrorMsg ( log_buff );

      return ( -1 );
    }
  }

  return ( -2 );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
int WavFirstBlock ( int fd, char * stype )
{
  int           rc;

  /* Go back to the beginning of the file (12 bytes into it, allowing for   */
  /* the 'RIFF' + long_length + 'WAVE' header info                          */

  if ( ( rc = lseek ( fd, 12, SEEK_SET ) ) == -1 )
  {
    sprintf ( log_buff, "WavFirstBlock: lseek failed, errno = %d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }


  /* Now just go and find the 'next' instance of the block we are        */
  /* interested in...                                                    */

  return ( WavNextBlock ( fd, stype ) );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
int WavDataBlockSize ( int fd )
{
  unsigned long ltemp;

  wav_data_pos = 0;

  if ( NextLong ( fd, & ltemp ) )
  {
    sprintf ( log_buff, "WavDataBlockSize: read(len) failed, errno = %d", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  wav_data_len = SwapLong ( ltemp );

  return ( 0 );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
int WavNextDataBlock ( int fd )
{
  if ( WavNextBlock ( fd, "data" ) )
  {
    sprintf ( log_buff, "WavNextDataBlock: No next block found\n" );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  return ( WavDataBlockSize ( fd ) );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
int WavFirstDataBlock ( int fd )
{
  if ( WavFirstBlock ( fd, "data" ) )
  {
    sprintf ( log_buff, "WavFirstDataBlock: No first block found\n" );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  return ( WavDataBlockSize ( fd ) );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
int WavSkipBlock ( int fd )
{
  int           rc;
  unsigned long ltemp;
  long          block_len;

  /*-----------------------------------------------------------------------*/
  /* This function assumes that the file pointer is pointing to the length */
  /* of the block that is about to be skipped!!!!!                         */
  /*-----------------------------------------------------------------------*/

  if ( NextLong ( fd, & ltemp ) )
  {
    sprintf ( log_buff, "WavSkipBlock: read(length) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  block_len = SwapLong ( ltemp );


  if ( ( rc = lseek ( fd, block_len, SEEK_CUR ) ) == -1 )
  {
    sprintf ( log_buff, "WavSkipBlock: lseek failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }

  return ( 0 );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
int WavOpenFile ( char * fn )
{
  int           fd;
  unsigned long ltemp;
  char          stemp [ 4 ];

  if ( ! ( fd = open ( fn, O_RDONLY | O_BINARY ) ) )
  {
    sprintf ( log_buff, "WavOpenFile: open failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    return ( -1 );
  }


  if ( NextLong ( fd, ( unsigned long * ) stemp ) )
  {
    sprintf ( log_buff, "WavOpenFile: read(RIFF) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    close ( fd );

    return ( -1 );
  }

  if ( strncmp ( stemp, "RIFF", 4 ) )
  {
    sprintf ( log_buff, "WavOpenFile: non-RIFF format\n" );
    ErrorMsg ( log_buff );

    close ( fd );

    return ( -1 );
  }


  if ( NextLong ( fd, & ltemp ) )
  {
    sprintf ( log_buff, "WavOpenFile: read(length) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    close ( fd );

    return ( -1 );
  }



  if ( NextLong ( fd, ( unsigned long * ) stemp ) )
  {
    sprintf ( log_buff, "WavOpenFile: read(WAVE) failed, errno=%d\n", errno );
    ErrorMsg ( log_buff );

    close ( fd );

    return ( -1 );
  }

  if ( strncmp ( stemp, "WAVE", 4 ) )
  {
    sprintf ( log_buff, "WavOpenFile: non-WAVE format\n" );
    ErrorMsg ( log_buff );

    close ( fd );

    return ( -1 );
  }

  return ( fd );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
int WavCloseFile ( int fd )
{
  close ( fd );

  return ( 0 );
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/*--------------------------------------------------------------------------*/
int WavReadBlock ( int fd, char * buff, int siz )
{
  int bytes_needed;
  int bytes_copied;
  int this_chunk;
  int rc;

  /* If there's enough data in the buffer, read whatever is needed and */
  /* return it to the caller (adjusting pointers, of course)           */
  if ( siz <= ( wav_data_len - wav_data_pos ) )
  {
    rc = read ( fd, buff, siz );

    wav_data_pos = wav_data_pos + siz;

    return ( siz );
  }

  /* Guess there wasn't enough stuff in the current data block to */
  /* satisfy the outstanding request, so we copy whatever we have */
  /* and then try to find another block of data and get stuff out */
  /* of that one...                                               */
  bytes_copied = wav_data_len - wav_data_pos;

  rc = read ( fd, buff, bytes_copied );

  bytes_needed = siz - bytes_copied;

  while ( bytes_copied < siz )
  {
    if ( WavNextDataBlock ( fd ) )
    {
      sprintf ( log_buff, "WavReadBlock: no more data\n" );
      LogMsg ( log_buff );

      return ( bytes_copied );
    }

    /* If there's now enough data to fulfill the rest of our request, */
    /* then we read it (and weep - ha ha ha), and return ...          */
    if ( ( wav_data_len - wav_data_pos ) > bytes_needed )
    {
      rc = read ( fd, & buff [ bytes_copied ], bytes_needed );

      wav_data_pos = wav_data_pos + bytes_needed;

      return ( siz );
    }

    /* There wasn't enough data in the newly acquired buffer, so we */
    /* take what we can get and iterate..                           */
    this_chunk = wav_data_len - wav_data_pos;

    rc = read ( fd, & buff [ bytes_copied ], this_chunk );

    bytes_copied = bytes_copied + this_chunk;

    wav_data_pos = wav_data_pos + this_chunk;

    bytes_needed = bytes_needed - this_chunk;
  }
}

//--------------------------------------------------------------------------
//  WavCheckFormat
//
//  Return Codes:
//       0 - Format is correct
//    -900 - Unable to open file
//    -901 - Invalid Wavefile format 
//    -902 - Vocabulary not campatable with wavefile format
//                                                                      
//--------------------------------------------------------------------------
int WavCheckFormat( char *szWaveFile, int iSampleRate )
{
   int            iRC = -901;
   HMMIO          hmmio;
   MMCKINFO       mmckinfoParent;
   MMCKINFO       mmckinfoSubchunk;
   DWORD          dwFmtSize;
   WAVEFORMATEX   waveFormat;

   if (hmmio = mmioOpen( szWaveFile, NULL, MMIO_READ | MMIO_ALLOCBUF))
   {
      mmckinfoParent.fccType = mmioFOURCC('W','A','V','E');
      if (mmioDescend(hmmio, (LPMMCKINFO)&mmckinfoParent, NULL, MMIO_FINDRIFF)==MMSYSERR_NOERROR)
      {
         mmckinfoSubchunk.ckid = mmioFOURCC('f','m','t',' ');
         if (mmioDescend(hmmio, (LPMMCKINFO)&mmckinfoSubchunk, &mmckinfoParent, MMIO_FINDCHUNK)==MMSYSERR_NOERROR)
         {
            dwFmtSize = mmckinfoSubchunk.cksize;
            if (mmioRead(hmmio, (HPSTR)&waveFormat, dwFmtSize) == (LONG)dwFmtSize)
            {
               // Check the format
               if ( (waveFormat.wFormatTag     == WAVE_FORMAT_PCM) &&
                    (waveFormat.nChannels      == 1)               &&
                    (waveFormat.wBitsPerSample == 16)              )
               {
                  if (waveFormat.nSamplesPerSec != 11025)
                     iRC = -901;
                  else if (waveFormat.nSamplesPerSec != (UINT)iSampleRate)
                     iRC = -902;
                  else
                     iRC = 0;
               }
               else
                  iRC = -901;
            }
         }
      }
      mmioClose(hmmio, 0);
   }
   else
      iRC = -900;

   return iRC;
}

